pub mod color;
pub mod font;

use std::fmt::Display;

use color::ColorPair;
use font::Font;
use wmenu_rs::{Menu as WMenu, MenuError};

#[derive(Debug)]
pub enum Error {
    WMenu(MenuError),
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Error::WMenu(menu_error) => write!(f, "{menu_error}"),
        }
    }
}

impl std::error::Error for Error {}

impl From<MenuError> for Error {
    fn from(value: MenuError) -> Self {
        Self::WMenu(value)
    }
}

pub struct Config {
    pub bottom: bool,
    pub padding: usize,
    pub lines: i32,
    pub font_family: String,
    pub font_size: usize,
    pub language: Option<String>,
    pub normal_color: ColorPair,
    pub prompt_color: ColorPair,
    pub selection_color: ColorPair,
}

impl Config {
    fn load_font(&self) -> Font {
        Font::load_from_string(self.font_family.as_str(), self.font_size).unwrap_or_default()
    }

    fn line_height(&self, font_height: i32) -> i32 {
        font_height + 2
    }

    fn height(&self, line_height: i32) -> i32 {
        if self.lines > 0 {
            line_height + line_height * self.lines
        } else {
            line_height
        }
    }

    fn apply(self, wmenu: &mut WMenu) {
        let font = self.load_font();

        let font_height = font.height();
        let line_height = self.line_height(font_height);
        let height = self.height(line_height);

        wmenu
            .height(height)
            .bottom(self.bottom)
            .padding(self.padding.try_into().unwrap())
            .lines(self.lines.try_into().unwrap())
            .line_height(line_height)
            .font(&font.to_string())
            .normal_bg(self.normal_color.bg.into())
            .normal_fg(self.normal_color.fg.into())
            .prompt_bg(self.prompt_color.bg.into())
            .prompt_fg(self.prompt_color.fg.into())
            .selection_bg(self.selection_color.bg.into())
            .selection_fg(self.selection_color.fg.into());
    }
}

impl Default for Config {
    fn default() -> Self {
        Self {
            bottom: false,
            lines: 0,
            padding: 16,
            font_family: "monospace".to_string(),
            font_size: 12,
            language: None,
            normal_color: (0x222222ff, 0xbbbbbbffu32).into(),
            prompt_color: (0x005577ff, 0xeeeeeeffu32).into(),
            selection_color: (0x005577ff, 0x227799ffu32).into(),
        }
    }
}

pub struct Menu {
    wmenu: WMenu,
}

impl Menu {
    pub fn new(config: Config) -> Self {
        let mut wmenu = WMenu::new();
        config.apply(&mut wmenu);
        Self { wmenu }
    }

    pub fn inner(&self) -> &WMenu {
        &self.wmenu
    }

    pub fn inner_mut(&mut self) -> &mut WMenu {
        &mut self.wmenu
    }

    pub fn add_item(&mut self, item: &str) -> &mut Menu {
        self.inner_mut().add_item(item);
        self
    }

    pub fn add_items(&mut self, items: &[&str]) -> &mut Menu {
        self.inner_mut().add_items(items);
        self
    }

    pub fn run(&self) -> Result<String, Error> {
        Ok(self.inner().run()?)
    }
}

impl Default for Menu {
    fn default() -> Self {
        Self::new(Config::default())
    }
}

impl From<Menu> for WMenu {
    fn from(value: Menu) -> Self {
        value.wmenu
    }
}

#[cfg(test)]
mod tests {
    use crate::{Config, Menu};

    #[test]
    fn it_works() {
        let items = ["rowern", "sybil", "vex"];
        let mut menu = Menu::new(Config::default());
        menu.add_items(&items);
        let result = menu.run();
        assert!(items.contains(&result.unwrap().as_str()));
    }
}
